import os
import random
from PIL import Image
import numpy as np

# Creates the Mini-ImageNet directory

# Some useful variables
seed = 0.1
source_dir = '../../media/data/train/'
destination_dir = './Mini-ImageNet/'
num_classes = 100
image_size = (84, 84)
num_images_per_class = 600


def load_image(filename, image_size):
	"""
		load_image(str) -> Image
		Loads the image file and returns a 3 channel image of size = image_size
	"""
	# Read the image
	im = Image.open(filename)
	im.load()
	
	# Crop the image
	w = im.size[0]
	h = im.size[1]
	if h < w:
		im = im.crop(((w-h)/2, 0, (w-h)/2 + h, h))
	else:
		im = im.crop((0, (h-w)/2, w, (h-w)/2 + w))
	
	# Resize the image
	im = im.resize(image_size, Image.LANCZOS)
	
	# Convert the image to an ndaaray
	im = np.asarray(im)
	shape = np.shape(im)
	if len(shape) == 2:	# Only 1 channel
		im = np.stack([im, im, im], axis=2)	# Convert to a 3 channel image
	elif shape[2] == 4:	# 4 Channels
		im = im[:, :, 1:4]	# Keep the RGB channels
	
	im = Image.fromarray(im)
			
	return im


# Set random seed
random.seed(seed)

############################################ Choose Random Classes #################################################
# Read all the classes
dirnames = []
for _, dirnames, _ in os.walk(source_dir):
	break

classes = []
for dirname in dirnames:
	if dirname[0] == 'n':	# Class folder start with n#id
		classes.append(dirname)

# Choose num_classes classes randomly
classes = random.sample(classes, num_classes)

############################################### Read the classes ###################################################
class_num = 0
for class_dir in classes:
	class_num += 1
	num_files = 0
	print 'Reading Class:', (class_num), 'Progress:', (num_files/float(num_images_per_class))*100,
	# Read the image file names in this class
	filenames = []
	for _, _, filenames in os.walk(os.path.join(source_dir, class_dir)):
		break
	
	# Randomly choose num_image_per_class
	filenames = random.sample(filenames, num_images_per_class)
	
	# Create the destination directory
	os.mkdir(os.path.join(destination_dir, class_dir))
	
	# Write the files to destination directory
	for filename in filenames:
		im = load_image(os.path.join(source_dir, class_dir, filename), image_size)
		im.save(os.path.join(destination_dir, class_dir, filename))
		num_files += 1
		print '\rReading Class:', (class_num), 'Progress:', (num_files/float(num_images_per_class))*100,
	
	print ''

