import tensorflow as tf
import numpy as np
from read_mini_imagenet import *
from keras import backend as K
from keras.layers import Dense, Activation, Convolution2D, BatchNormalization, Dropout, Merge, LSTM
from keras.regularizers import l2
import math

# Useful variables
image_size = 32
attention_window_size = 4
num_glimpses = 8
batch_size = 128
learning_rate = 1e-4
hidden_layer_size = 256
num_batches = 100000
forget_bias = 0.2
wrn_k = 2
wrn_n = 4
reg_param = 0.00005

tf.reset_default_graph()
g1 = tf.Graph()

with g1.as_default() as g:
	def get_glimpse_batch(images, h, W_g):
		"""
			Get glimpse from the batch based on the hidden state
		"""
		omega = tf.nn.tanh(tf.matmul(W_g, h))
		x_hat = omega[0, :]
		y_hat = omega[1, :]
		delta_hat = omega[2, :]
		image_size = 16
		num_input_channels = 32*wrn_k
	
		x = tf.reshape((image_size - 1) * (x_hat + 1) / 2, (-1, 1))
		y = tf.reshape((image_size - 1) * (y_hat + 1) / 2, (-1, 1))
		delta = tf.reshape((image_size / float(attention_window_size)) * (1 - tf.abs(delta_hat)), (-1, 1))
		gamma = tf.reshape(tf.exp(1 - 2*tf.abs(delta_hat)), (-1, 1))

		grid_i = tf.reshape(tf.cast(tf.range(attention_window_size), tf.float32), [1, -1])
		ux = x + (grid_i - attention_window_size/2.0 - 0.5) * delta
		uy = y + (grid_i - attention_window_size/2.0 - 0.5) * delta

		a = tf.reshape(tf.cast(tf.range(image_size), tf.float32), [1, 1, -1])
		b = tf.reshape(tf.cast(tf.range(image_size), tf.float32), [1, 1, -1])
		ux = tf.reshape(ux, [-1, attention_window_size, 1])
		uy = tf.reshape(uy, [-1, attention_window_size, 1])
		pi = tf.constant(math.pi, dtype=tf.float32)

		gamma = tf.reshape(gamma, (-1, 1, 1))
		Fx = 1.0 / (tf.multiply(pi, gamma) * (1.0 + tf.square((a - ux) / gamma)))
		Fy = 1.0 / (tf.multiply(pi, gamma) * (1.0 + tf.square((b - uy) / gamma)))

		Fx = Fx / tf.reduce_sum(Fx, 2, keep_dims=True)
		Fy = Fy / tf.reduce_sum(Fy, 2, keep_dims=True)
	
	
		Fx = tf.stack([Fx]*num_input_channels, axis=1)
		Fy = tf.stack([Fy]*num_input_channels, axis=1)
	
		return tf.matmul(Fy, tf.matmul(tf.transpose(images, perm=[0, 3, 1, 2]), \
												tf.transpose(Fx, perm=[0, 1, 3, 2])))
											
											
											
	def residual_block(x, first=False, increase_dim=False, nb_filters=16):
		if increase_dim:
			first_stride = (2, 2)
		else:
			first_stride = (1, 1)
	
		if first:
			bn_pre_relu = x
		else:
			bn_pre_conv = BatchNormalization(epsilon=1e-4, momentum=0.9, mode=2)(x)
			bn_pre_relu = Activation('relu')(bn_pre_conv)
	
	
		# First layer
		l = Convolution2D(nb_filter=nb_filters, nb_row=3, nb_col=3, init='he_normal', border_mode='same', \
										subsample=first_stride, W_regularizer=l2(reg_param))(bn_pre_relu)
		l = BatchNormalization(epsilon=1e-4, momentum=0.9, mode=2)(l)
		l = Activation('relu')(l)
	
		# Dropout
		l = Dropout(0.3)(l)
	
		# Second layer
		conv2 = Convolution2D(nb_filter=nb_filters, nb_row=3, nb_col=3, init='he_normal', border_mode='same', \
										subsample=(1, 1), W_regularizer=l2(reg_param))(l)
	
	
		if increase_dim:
			proj = Convolution2D(nb_filter=nb_filters, nb_row=1, nb_col=1, init='he_normal', border_mode='same', \
										subsample=(1, 1), W_regularizer=l2(reg_param), bias=False)(conv2)
			l = conv2 + proj
		elif first:
			proj = Convolution2D(nb_filter=nb_filters, nb_row=1, nb_col=1, init='he_normal', border_mode='same', \
										subsample=(1, 1), W_regularizer=l2(reg_param), bias=False)(conv2)
			l = conv2 + proj
		else:
			l = conv2 + x
	
		return l



	inputs1 = tf.placeholder(dtype=tf.float32, shape=[batch_size, image_size, image_size, 3])
	inputs2 = tf.placeholder(dtype=tf.float32, shape=[batch_size, image_size, image_size, 3])
	labels = tf.placeholder(dtype=tf.float32, shape=[batch_size, 1])

	################################################ Run WideResNet ##################################################
	# Merge inputs1 and inputs2
	inputs = tf.concat([inputs1, inputs2], axis=0)

	n_filters = [16, 16*wrn_k, 32*wrn_k]

	# First layer
	l = Convolution2D(nb_filter=n_filters[0], nb_row=3, nb_col=3, init='he_normal', border_mode='same', \
										subsample=(1, 1), W_regularizer=l2(reg_param))(inputs)
	l = BatchNormalization(epsilon=1e-4, momentum=0.9, mode=2)(l)
	l = Activation('relu')(l)

	# Second layer, a stack of residual blocks
	l = residual_block(l, first=True, nb_filters=n_filters[1])
	for _ in range(1, wrn_n):
		l = residual_block(l, nb_filters=n_filters[1])

	# Third layer, a stack of residual blocks
	l = residual_block(l, increase_dim=True, nb_filters=n_filters[2])
	for _ in range(1, wrn_n+2):
		l = residual_block(l, nb_filters=n_filters[2])

	l = BatchNormalization(epsilon=1e-4, momentum=0.9, mode=2)(l)
	lstm_inputs = Activation('relu')(l)

	# Divide lstm_inputs
	lstm_inputs1 = lstm_inputs[:batch_size, :, :, :]
	lstm_inputs2 = lstm_inputs[batch_size:, :, :, :]



	############################################# Run the LSTM ########################################################
	lstm_cell = tf.contrib.rnn.BasicLSTMCell(hidden_layer_size, forget_bias = forget_bias)

	with tf.variable_scope("attention"):
		W_g = tf.Variable(tf.random_normal([3, hidden_layer_size], stddev=0.1))

	with tf.variable_scope("classifier"):
		W_c = tf.Variable(tf.random_normal([hidden_layer_size, 1], stddev=0.1))

	#"""
	h = lstm_cell.zero_state(batch_size, dtype=tf.float32)
	c = tf.zeros((batch_size, hidden_layer_size))
	for j in range(2*num_glimpses):
		if j % 2 == 0:
			glimpse = get_glimpse_batch(lstm_inputs1, tf.transpose(c), W_g)
			glimpse = tf.reshape(glimpse, [batch_size, -1])
		else:
			glimpse = get_glimpse_batch(lstm_inputs2, tf.transpose(c), W_g)
			glimpse = tf.reshape(glimpse, [batch_size, -1])
		with tf.variable_scope('mylstm') as scope:
			if j > 0:
				scope.reuse_variables()
			c, h = lstm_cell(glimpse, h)
		
	output = tf.nn.sigmoid(tf.matmul(c, W_c))
	network_cost = -tf.reduce_mean(labels*tf.log(output) + (1.0-labels)*tf.log(1.0-output))
	network_cost = network_cost + reg_param*tf.nn.l2_loss(W_c)# + reg_param*tf.nn.l2_loss(W_g)
	#"""

	train_step = tf.train.AdamOptimizer(learning_rate).minimize(network_cost)
	accuracy = tf.reduce_mean(tf.cast(tf.equal(tf.round(output), labels), tf.float32))


	def train(images1, images2, is_similar):
		"""
			Train the network for verification task
		"""
		_, batch_cost = sess.run((train_step, network_cost), feed_dict={inputs1: images1, inputs2: images2, labels: is_similar, K.learning_phase(): 1})
		return batch_cost
	
	
	def calculate_accuracy(images1, images2, is_similar, is_learning=0):
		"""
			Calculate accuracy for verification task
		"""
		#print sess.run(output, feed_dict={inputs1: images1, inputs2: images2, labels: is_similar})
		return sess.run(accuracy, feed_dict={inputs1: images1, inputs2: images2, labels: is_similar, K.learning_phase(): is_learning})
	
	

	sess = tf.InteractiveSession(graph=g1)
	saver = tf.train.Saver()
	if os.path.isfile('./Model/arc-model.meta'):
		saver.restore(sess, './Model/arc-model')
	else:
		sess.run(tf.global_variables_initializer())

	K.set_session(sess)
	
	
	
	
tf.reset_default_graph()
g2 = tf.Graph()
with g2.as_default() as g:
	g2_inputs = tf.placeholder(dtype=tf.float32, shape=[32, 5, hidden_layer_size])
	g2_labels = tf.placeholder(dtype=tf.float32, shape=[32, 5])
	
	g2_lstm = tf.contrib.rnn.BasicLSTMCell(128)
	
	# Go for an LSTM up pass
	g2_h_up = g2_lstm.zero_state(32, dtype=tf.float32)
	g2_c_up = tf.zeros((32, 128))
	for j in range(5):
		with tf.variable_scope('g2_mylstm') as scope:
			if j > 0:
				scope.reuse_variables()
			g2_c_up, g2_h_up = g2_lstm(tf.reshape(g2_inputs[:, j, :], [32, hidden_layer_size]), g2_h_up)
				
	# Go for an LSTM down pass
	g2_h_down = g2_lstm.zero_state(32, dtype=tf.float32)
	g2_c_down = tf.zeros((32, 128))
	for j in range(5):
		with tf.variable_scope('g2_mylstm') as scope:
			scope.reuse_variables()
			g2_c_down, g2_h_down = g2_lstm(tf.reshape(g2_inputs[:, 5-j-1, :], [32, hidden_layer_size]), g2_h_down)
	
	g2_c = tf.concat([g2_c_up, g2_c_down], axis=1)
	g2_W = tf.get_variable(name='final_w', shape=[256, 5], initializer=tf.contrib.layers.xavier_initializer())
	g2_output = tf.nn.softmax(tf.matmul(g2_c, g2_W))
	
	g2_network_cost = tf.losses.softmax_cross_entropy(g2_labels, g2_output)
	g2_train_step = tf.train.AdamOptimizer(learning_rate).minimize(g2_network_cost)
	g2_accuracy = tf.reduce_mean(tf.cast(tf.equal(tf.argmax(g2_output, 1), tf.argmax(g2_labels, 1)), tf.float32))
	
	g2_sess = tf.InteractiveSession(graph=g2)
	g2_saver = tf.train.Saver()
	if os.path.isfile('./ModelFC/arc-model.meta'):
		g2_saver.restore(g2_sess, './ModelFC/arc-model')
	else:
		g2_sess.run(tf.global_variables_initializer())
		
		

# Useful controls
display_step = 50
save_step = 1000


# Read the Data
casia = MiniImageNet('./Mini-ImageNet', (32, 32))

# Get one images from each of 20 test classes, this becomes the training set
im1, _, _ = casia.get_one_shot_data(casia.num_test_classes)

def get_one_shot_data(self, n):
	"""
		Returns n way one shot data
	"""
	# Initialize the data
	training_images = np.zeros((n, self.image_size[0], self.image_size[1], 3))
	test_image = np.zeros((self.image_size[0], self.image_size[1], 3))
	
	# Put one random example from each class into training_images
	# Select random classes
	classes = random.sample(range(self.num_test_classes), n)
	k = 0
	for i in classes:
		# Select a random example
		training_images[k, :, :, :] = im1[i, :, :, :]
		k += 1
		
	# Choose a random test_image
	class_idx = random.choice(classes)
	test_image = im1[class_idx, :, :, :]
	
	return (training_images, test_image, classes.index(class_idx))
	

#"""
t_cost = 0.0
for i in range(num_batches):
	ip = np.zeros((32, 5, hidden_layer_size))
	lbl = np.zeros((32, 5))
	for j in range(32):
		training_images, test_images, correct_idx = get_one_shot_data(casia, 5)
		training_images1, test_images1, is_similar1 = casia.get_data(123)
		test_images = np.reshape(test_images, [1, image_size, image_size, 3])
	
		# Append images to match original batch_size
		training_images = np.concatenate([training_images, training_images1], axis=0)
		test_images = np.repeat(test_images, 5, axis=0)
		test_images = np.concatenate([test_images, test_images1], axis=0)
		
		# Get the input for final lstm
		tf.reset_default_graph()
		with g1.as_default() as g:
			ip[j, :, :] = sess.run(c, feed_dict={inputs1: training_images, inputs2: test_images, K.learning_phase(): 0})[:5, :]
		lbl[j, correct_idx] = 1.0
	
	# Feed the input to the second lstm
	tf.reset_default_graph()
	with g2.as_default() as g:
		_, train_cost = g2_sess.run((g2_train_step, g2_network_cost), feed_dict={g2_inputs: ip, g2_labels: lbl})
	
	t_cost = t_cost + (1.0/(i+1))*(train_cost - t_cost)	
	
		
	if i % display_step == 0:
		# Fetch Test Images
		training_images, test_images, correct_idx = casia.get_one_shot_data(5)
		training_images1, test_images1, is_similar1 = casia.get_data(123)
		test_images = np.reshape(test_images, [1, image_size, image_size, 3])
	
		# Append images to match original batch_size
		training_images = np.concatenate([training_images, training_images1], axis=0)
		test_images = np.repeat(test_images, 5, axis=0)
		test_images = np.concatenate([test_images, test_images1], axis=0)
		 
		# Prepare input for final lstm	
		tf.reset_default_graph()
		with g1.as_default() as g:
			ip = np.repeat(np.reshape(sess.run(c, feed_dict={inputs1: training_images, inputs2: test_images, K.learning_phase(): 0})[:5, :], (1, 5, 256)), 32, axis=0)
		lbl = np.zeros((32, 5))
		lbl[:, correct_idx] = 1.0
		
		tf.reset_default_graph()
		with g2.as_default() as g:
			test_acc = g2_sess.run(g2_accuracy, feed_dict={g2_inputs: ip, g2_labels: lbl})
			print g2_sess.run(g2_output, feed_dict={g2_inputs: ip, g2_labels: lbl})[0, :]
			print correct_idx
		
		print "Batch:", (i+1), "Train Cost:", t_cost, "Test Accuracy:", test_acc
		
	if i % save_step == 0:
		g2_saver.save(g2_sess, './ModelFC/arc-model')
#"""

# Run 5-way one-shot classification
"""
num_trials = 10000
num_correct = 0
for i in range(num_trials):
	training_images, test_images, correct_idx = casia.get_one_shot_data(5)
	training_images1, test_images1, is_similar1 = casia.get_data(123)
	test_images = np.reshape(test_images, [1, image_size, image_size, 3])
	
	# Append images to match original batch_size
	training_images = np.concatenate([training_images, training_images1], axis=0)
	test_images = np.repeat(test_images, 5, axis=0)
	test_images = np.concatenate([test_images, test_images1], axis=0)
	
	# Get the network output
	network_output = sess.run(output, feed_dict={inputs1: training_images, inputs2: test_images, K.learning_phase(): 0})
	network_output = network_output[:5]	# Remove all padded images
	predicted_idx = np.argmax(network_output.flatten())
	if predicted_idx == correct_idx:
		num_correct += 1
	print predicted_idx == correct_idx
	
print num_correct / float(num_trials)
"""
